@@ -12,7 +12,6 @@
 #include "postgres_fe.h"
 
 #include "dumputils.h"
-#include "fe_utils/connect.h"
 #include "fe_utils/string_utils.h"
 #include "parallel.h"
 #include "pg_backup_archiver.h"
@@ -26,12 +25,11 @@
 #endif
 
 
-#define DB_MAX_ERR_STMT 128
-
 /* translator: this is a module name */
 static const char *modulename = gettext_noop("archiver (db)");
 
 static void _check_database_version(ArchiveHandle *AH);
+static PGconn *_connectDB(ArchiveHandle *AH, const char *newdbname, const char *newUser);
 static void notice_processor(void *arg, const char *message);
 
 static void
@@ -77,68 +75,197 @@ _check_database_version(ArchiveHandle *AH)
 
 /*
  * Reconnect to the server.  If dbname is not NULL, use that database,
- * else the one associated with the archive handle.
+ * else the one associated with the archive handle.  If username is
+ * not NULL, use that user name, else the one from the handle.  If
+ * both the database and the user match the existing connection already,
+ * nothing will be done.
+ *
+ * Returns 1 in any case.
  */
-void
-ReconnectToServer(ArchiveHandle *AH, const char *dbname)
+int
+ReconnectToServer(ArchiveHandle *AH, const char *dbname, const char *username)
 {
-	PGconn	   *oldConn = AH->connection;
-	RestoreOptions *ropt = AH->public.ropt;
+	PGconn	   *newConn;
+	const char *newdbname;
+	const char *newusername;
 
-	/*
-	 * Save the dbname, if given, in override_dbname so that it will also
-	 * affect any later reconnection attempt.
-	 */
-	if (dbname)
-		ropt->cparams.override_dbname = pg_strdup(dbname);
+	if (!dbname)
+		newdbname = PQdb(AH->connection);
+	else
+		newdbname = dbname;
 
-	/*
-	 * Note: we want to establish the new connection, and in particular update
-	 * ArchiveHandle's connCancel, before closing old connection.  Otherwise
-	 * an ill-timed SIGINT could try to access a dead connection.
-	 */
-	AH->connection = NULL;		/* dodge error check in ConnectDatabase */
+	if (!username)
+		newusername = PQuser(AH->connection);
+	else
+		newusername = username;
 
-	ConnectDatabase((Archive *) AH, &ropt->cparams, true);
+	/* Let's see if the request is already satisfied */
+	if (strcmp(newdbname, PQdb(AH->connection)) == 0 &&
+		strcmp(newusername, PQuser(AH->connection)) == 0)
+		return 1;
 
-	PQfinish(oldConn);
+	newConn = _connectDB(AH, newdbname, newusername);
+
+	/* Update ArchiveHandle's connCancel before closing old connection */
+	set_archive_cancel_info(AH, newConn);
+
+	PQfinish(AH->connection);
+	AH->connection = newConn;
+
+	return 1;
 }
 
 /*
- * Make, or remake, a database connection with the given parameters.
- *
- * The resulting connection handle is stored in AHX->connection.
+ * Connect to the db again.
  *
+ * Note: it's not really all that sensible to use a single-entry password
+ * cache if the username keeps changing.  In current usage, however, the
+ * username never does change, so one savedPassword is sufficient.  We do
+ * update the cache on the off chance that the password has changed since the
+ * start of the run.
+ */
+static PGconn *
+_connectDB(ArchiveHandle *AH, const char *reqdb, const char *requser)
+{
+	PQExpBufferData connstr;
+	PGconn	   *newConn;
+	const char *newdb;
+	const char *newuser;
+	char	   *password;
+	char		passbuf[100];
+	bool		new_pass;
+
+	if (!reqdb)
+		newdb = PQdb(AH->connection);
+	else
+		newdb = reqdb;
+
+	if (!requser || strlen(requser) == 0)
+		newuser = PQuser(AH->connection);
+	else
+		newuser = requser;
+
+	ahlog(AH, 1, "connecting to database \"%s\" as user \"%s\"\n",
+		  newdb, newuser);
+
+	password = AH->savedPassword;
+
+	if (AH->promptPassword == TRI_YES && password == NULL)
+	{
+		simple_prompt("Password: ", passbuf, sizeof(passbuf), false);
+		password = passbuf;
+	}
+
+	initPQExpBuffer(&connstr);
+	appendPQExpBuffer(&connstr, "dbname=");
+	appendConnStrVal(&connstr, newdb);
+
+	do
+	{
+		const char *keywords[7];
+		const char *values[7];
+
+		keywords[0] = "host";
+		values[0] = PQhost(AH->connection);
+		keywords[1] = "port";
+		values[1] = PQport(AH->connection);
+		keywords[2] = "user";
+		values[2] = newuser;
+		keywords[3] = "password";
+		values[3] = password;
+		keywords[4] = "dbname";
+		values[4] = connstr.data;
+		keywords[5] = "fallback_application_name";
+		values[5] = progname;
+		keywords[6] = NULL;
+		values[6] = NULL;
+
+		new_pass = false;
+		newConn = PQconnectdbParams(keywords, values, true);
+
+		if (!newConn)
+			exit_horribly(modulename, "failed to reconnect to database\n");
+
+		if (PQstatus(newConn) == CONNECTION_BAD)
+		{
+			if (!PQconnectionNeedsPassword(newConn))
+				exit_horribly(modulename, "could not reconnect to database: %s",
+							  PQerrorMessage(newConn));
+			PQfinish(newConn);
+
+			if (password)
+				fprintf(stderr, "Password incorrect\n");
+
+			fprintf(stderr, "Connecting to %s as %s\n",
+					newdb, newuser);
+
+			if (AH->promptPassword != TRI_NO)
+			{
+				simple_prompt("Password: ", passbuf, sizeof(passbuf), false);
+				password = passbuf;
+			}
+			else
+				exit_horribly(modulename, "connection needs password\n");
+
+			new_pass = true;
+		}
+	} while (new_pass);
+
+	/*
+	 * We want to remember connection's actual password, whether or not we got
+	 * it by prompting.  So we don't just store the password variable.
+	 */
+	if (PQconnectionUsedPassword(newConn))
+	{
+		if (AH->savedPassword)
+			free(AH->savedPassword);
+		AH->savedPassword = pg_strdup(PQpass(newConn));
+	}
+
+	termPQExpBuffer(&connstr);
+
+	/* check for version mismatch */
+	_check_database_version(AH);
+
+	PQsetNoticeProcessor(newConn, notice_processor, NULL);
+
+	return newConn;
+}
+
+
+/*
+ * Make a database connection with the given parameters.  The
+ * connection handle is returned, the parameters are stored in AHX.
  * An interactive password prompt is automatically issued if required.
- * We store the results of that in AHX->savedPassword.
+ *
  * Note: it's not really all that sensible to use a single-entry password
  * cache if the username keeps changing.  In current usage, however, the
  * username never does change, so one savedPassword is sufficient.
  */
 void
 ConnectDatabase(Archive *AHX,
-				const ConnParams *cparams,
-				bool isReconnect)
+				const char *dbname,
+				const char *pghost,
+				const char *pgport,
+				const char *username,
+				trivalue prompt_password)
 {
 	ArchiveHandle *AH = (ArchiveHandle *) AHX;
-	trivalue	prompt_password;
 	char	   *password;
+	char		passbuf[100];
 	bool		new_pass;
 
 	if (AH->connection)
 		exit_horribly(modulename, "already connected to a database\n");
 
-	/* Never prompt for a password during a reconnection */
-	prompt_password = isReconnect ? TRI_NO : cparams->promptPassword;
-
-	password = AH->savedPassword ? pg_strdup(AH->savedPassword) : NULL;
+	password = AH->savedPassword;
 
 	if (prompt_password == TRI_YES && password == NULL)
 	{
-		password = simple_prompt("Password: ", 100, false);
-		if (password == NULL)
-			exit_horribly(modulename, "out of memory\n");
+		simple_prompt("Password: ", passbuf, sizeof(passbuf), false);
+		password = passbuf;
 	}
+	AH->promptPassword = prompt_password;
 
 	/*
 	 * Start the connection.  Loop until we have a password if requested by
@@ -146,35 +273,23 @@ ConnectDatabase(Archive *AHX,
 	 */
 	do
 	{
-		const char *keywords[8];
-		const char *values[8];
-		int			i = 0;
+		const char *keywords[7];
+		const char *values[7];
 
-		/*
-		 * If dbname is a connstring, its entries can override the other
-		 * values obtained from cparams; but in turn, override_dbname can
-		 * override the dbname component of it.
-		 */
-		keywords[i] = "host";
-		values[i++] = cparams->pghost;
-		keywords[i] = "port";
-		values[i++] = cparams->pgport;
-		keywords[i] = "user";
-		values[i++] = cparams->username;
-		keywords[i] = "password";
-		values[i++] = password;
-		keywords[i] = "dbname";
-		values[i++] = cparams->dbname;
-		if (cparams->override_dbname)
-		{
-			keywords[i] = "dbname";
-			values[i++] = cparams->override_dbname;
-		}
-		keywords[i] = "fallback_application_name";
-		values[i++] = progname;
-		keywords[i] = NULL;
-		values[i++] = NULL;
-		Assert(i <= lengthof(keywords));
+		keywords[0] = "host";
+		values[0] = pghost;
+		keywords[1] = "port";
+		values[1] = pgport;
+		keywords[2] = "user";
+		values[2] = username;
+		keywords[3] = "password";
+		values[3] = password;
+		keywords[4] = "dbname";
+		values[4] = dbname;
+		keywords[5] = "fallback_application_name";
+		values[5] = progname;
+		keywords[6] = NULL;
+		values[6] = NULL;
 
 		new_pass = false;
 		AH->connection = PQconnectdbParams(keywords, values, true);
@@ -188,30 +303,17 @@ ConnectDatabase(Archive *AHX,
 			prompt_password != TRI_NO)
 		{
 			PQfinish(AH->connection);
-			password = simple_prompt("Password: ", 100, false);
-			if (password == NULL)
-				exit_horribly(modulename, "out of memory\n");
+			simple_prompt("Password: ", passbuf, sizeof(passbuf), false);
+			password = passbuf;
 			new_pass = true;
 		}
 	} while (new_pass);
 
 	/* check to see that the backend connection was successfully made */
 	if (PQstatus(AH->connection) == CONNECTION_BAD)
-	{
-		if (isReconnect)
-			exit_horribly(modulename, "reconnection to database \"%s\" failed: %s",
-						  PQdb(AH->connection) ? PQdb(AH->connection) : "",
-						  PQerrorMessage(AH->connection));
-		else
-			exit_horribly(modulename, "connection to database \"%s\" failed: %s",
-						  PQdb(AH->connection) ? PQdb(AH->connection) : "",
-						  PQerrorMessage(AH->connection));
-	}
-
-	/* Start strict; later phases may override this. */
-	if (PQserverVersion(AH->connection) >= 70300)
-		PQclear(ExecuteSqlQueryForSingleRow((Archive *) AH,
-											ALWAYS_SECURE_SEARCH_PATH_SQL));
+		exit_horribly(modulename, "connection to database \"%s\" failed: %s",
+					  PQdb(AH->connection) ? PQdb(AH->connection) : "",
+					  PQerrorMessage(AH->connection));
 
 	/*
 	 * We want to remember connection's actual password, whether or not we got
@@ -223,8 +325,6 @@ ConnectDatabase(Archive *AHX,
 			free(AH->savedPassword);
 		AH->savedPassword = pg_strdup(PQpass(AH->connection));
 	}
-	if (password)
-		free(password);
 
 	/* check for version mismatch */
 	_check_database_version(AH);
@@ -347,7 +447,6 @@ ExecuteSqlCommand(ArchiveHandle *AH, const char *qry, const char *desc)
 {
 	PGconn	   *conn = AH->connection;
 	PGresult   *res;
-	char		errStmt[DB_MAX_ERR_STMT];
 
 #ifdef NOT_USED
 	fprintf(stderr, "Executing: '%s'\n\n", qry);
@@ -367,16 +466,8 @@ ExecuteSqlCommand(ArchiveHandle *AH, const char *qry, const char *desc)
 			break;
 		default:
 			/* trouble */
-			strncpy(errStmt, qry, DB_MAX_ERR_STMT);		/* strncpy required here */
-			if (errStmt[DB_MAX_ERR_STMT - 1] != '\0')
-			{
-				errStmt[DB_MAX_ERR_STMT - 4] = '.';
-				errStmt[DB_MAX_ERR_STMT - 3] = '.';
-				errStmt[DB_MAX_ERR_STMT - 2] = '.';
-				errStmt[DB_MAX_ERR_STMT - 1] = '\0';
-			}
 			warn_or_exit_horribly(AH, modulename, "%s: %s    Command was: %s\n",
-								  desc, PQerrorMessage(conn), errStmt);
+								  desc, PQerrorMessage(conn), qry);
 			break;
 	}
 
@@ -540,7 +631,7 @@ EndDBCopyMode(Archive *AHX, const char *tocEntryTag)
 		res = PQgetResult(AH->connection);
 		if (PQresultStatus(res) != PGRES_COMMAND_OK)
 			warn_or_exit_horribly(AH, modulename, "COPY failed for table \"%s\": %s",
-								tocEntryTag, PQerrorMessage(AH->connection));
+								  tocEntryTag, PQerrorMessage(AH->connection));
 		PQclear(res);
 
 		/* Do this to ensure we've pumped libpq back to idle state */
