@@ -43,7 +43,7 @@ PostgresNode - class representing PostgreSQL server instance
   # run query every second until it returns 't'
   # or times out
   $node->poll_query_until('postgres', q|SELECT random() < 0.1;|')
-    or print "timed out";
+    or die "timed out";
 
   # Do an online pg_basebackup
   my $ret = $node->backup('testbackup1');
@@ -93,16 +93,17 @@ use RecursiveCopy;
 use Socket;
 use Test::More;
 use TestLib ();
+use Time::HiRes qw(usleep);
 use Scalar::Util qw(blessed);
 
 our @EXPORT = qw(
   get_new_node
 );
 
-our ($use_tcp, $test_localhost, $test_pghost, $last_host_assigned,
-	$last_port_assigned, @all_nodes);
+our ($test_localhost, $test_pghost, $last_port_assigned, @all_nodes);
+
+# Windows path to virtual file system root
 
-# For backward compatibility only.
 our $vfs_path = '';
 if ($Config{osname} eq 'msys')
 {
@@ -113,14 +114,13 @@ if ($Config{osname} eq 'msys')
 INIT
 {
 
-	# Set PGHOST for backward compatibility.  This doesn't work for own_host
-	# nodes, so prefer to not rely on this when writing new tests.
-	$use_tcp            = $TestLib::windows_os;
-	$test_localhost     = "127.0.0.1";
-	$last_host_assigned = 1;
-	$test_pghost        = $use_tcp ? $test_localhost : TestLib::tempdir_short;
-	$ENV{PGHOST}        = $test_pghost;
-	$ENV{PGDATABASE}    = 'postgres';
+	# PGHOST is set once and for all through a single series of tests when
+	# this module is loaded.
+	$test_localhost = "127.0.0.1";
+	$test_pghost =
+	  $TestLib::windows_os ? $test_localhost : TestLib::tempdir_short;
+	$ENV{PGHOST}     = $test_pghost;
+	$ENV{PGDATABASE} = 'postgres';
 
 	# Tracking of last port value assigned to accelerate free port lookup.
 	$last_port_assigned = int(rand() * 16384) + 49152;
@@ -151,10 +151,7 @@ sub new
 		_host    => $pghost,
 		_basedir => TestLib::tempdir("data_" . $name),
 		_name    => $name,
-		_logfile_generation => 0,
-		_logfile_base       => "$TestLib::log_path/${testname}_${name}",
-		_logfile            => "$TestLib::log_path/${testname}_${name}.log"
-	};
+		_logfile => "$TestLib::log_path/${testname}_${name}.log" };
 
 	bless $self, $class;
 	$self->dump_info;
@@ -256,7 +253,13 @@ sub connstr
 	{
 		return "port=$pgport host=$pghost";
 	}
-	return "port=$pgport host=$pghost dbname=$dbname";
+
+	# Escape properly the database string before using it, only
+	# single quotes and backslashes need to be treated this way.
+	$dbname =~ s#\\#\\\\#g;
+	$dbname =~ s#\'#\\\'#g;
+
+	return "port=$pgport host=$pghost dbname='$dbname'";
 }
 
 =pod
@@ -354,13 +357,9 @@ sub set_replication_conf
 	$self->host eq $test_pghost
 	  or die "set_replication_conf only works with the default host";
 
-	open my $hba, ">>$pgdata/pg_hba.conf";
+	open my $hba, '>>', "$pgdata/pg_hba.conf";
 	print $hba "\n# Allow replication (set up by PostgresNode.pm)\n";
-	if (!$TestLib::windows_os)
-	{
-		print $hba "local replication all trust\n";
-	}
-	else
+	if ($TestLib::windows_os)
 	{
 		print $hba
 "host replication all $test_localhost/32 sspi include_realm=1 map=regress\n";
@@ -380,14 +379,13 @@ a directory that's only accessible to the current user to ensure that.
 On Windows, we use SSPI authentication to ensure the same (by pg_regress
 --config-auth).
 
-pg_hba.conf is configured to allow replication connections. Pass the keyword
-parameter hba_permit_replication => 0 to disable this.
-
 WAL archiving can be enabled on this node by passing the keyword parameter
 has_archiving => 1. This is disabled by default.
 
 postgresql.conf can be set up for replication by passing the keyword
-parameter allows_streaming => 1. This is disabled by default.
+parameter allows_streaming => 'logical' or 'physical' (passing 1 will also
+suffice for physical replication) depending on type of replication that
+should be enabled. This is disabled by default.
 
 The new node is set up in a fast but unsafe configuration where fsync is
 disabled.
@@ -401,39 +399,37 @@ sub init
 	my $pgdata = $self->data_dir;
 	my $host   = $self->host;
 
-	$params{hba_permit_replication} = 1
-	  unless defined $params{hba_permit_replication};
 	$params{allows_streaming} = 0 unless defined $params{allows_streaming};
 	$params{has_archiving}    = 0 unless defined $params{has_archiving};
 
 	mkdir $self->backup_dir;
 	mkdir $self->archive_dir;
 
-	TestLib::system_or_bail('initdb', '-D', $pgdata, '-A', 'trust', '-N');
+	TestLib::system_or_bail('initdb', '-D', $pgdata, '-A', 'trust', '-N',
+		@{ $params{extra} });
 	TestLib::system_or_bail($ENV{PG_REGRESS}, '--config-auth', $pgdata);
 
-	open my $conf, ">>$pgdata/postgresql.conf";
+	open my $conf, '>>', "$pgdata/postgresql.conf";
 	print $conf "\n# Added by PostgresNode.pm\n";
 	print $conf "fsync = off\n";
 	print $conf "restart_after_crash = off\n";
+	print $conf "log_line_prefix = '%m [%p] %q%a '\n";
 	print $conf "log_statement = all\n";
 	print $conf "wal_retrieve_retry_interval = '500ms'\n";
-
-	# If a setting tends to affect whether tests pass or fail, print it after
-	# TEMP_CONFIG.  Otherwise, print it before TEMP_CONFIG, thereby permitting
-	# overrides.  Settings that merely improve performance or ease debugging
-	# belong before TEMP_CONFIG.
-	print $conf TestLib::slurp_file($ENV{TEMP_CONFIG})
-	  if defined $ENV{TEMP_CONFIG};
-
-	# XXX Neutralize any stats_temp_directory in TEMP_CONFIG.  Nodes running
-	# concurrently must not share a stats_temp_directory.
-	print $conf "stats_temp_directory = 'pg_stat_tmp'\n";
+	print $conf "port = $port\n";
 
 	if ($params{allows_streaming})
 	{
-		print $conf "wal_level = replica\n";
+		if ($params{allows_streaming} eq "logical")
+		{
+			print $conf "wal_level = logical\n";
+		}
+		else
+		{
+			print $conf "wal_level = replica\n";
+		}
 		print $conf "max_wal_senders = 5\n";
+		print $conf "max_replication_slots = 5\n";
 		print $conf "wal_keep_segments = 20\n";
 		print $conf "max_wal_size = 128MB\n";
 		print $conf "shared_buffers = 1MB\n";
@@ -441,11 +437,14 @@ sub init
 		print $conf "hot_standby = on\n";
 		print $conf "max_connections = 10\n";
 	}
+	else
+	{
+		print $conf "wal_level = minimal\n";
+		print $conf "max_wal_senders = 0\n";
+	}
 
-	print $conf "port = $port\n";
-	if ($use_tcp)
+	if ($TestLib::windows_os)
 	{
-		print $conf "unix_socket_directories = ''\n";
 		print $conf "listen_addresses = '$host'\n";
 	}
 	else
@@ -455,7 +454,7 @@ sub init
 	}
 	close $conf;
 
-	$self->set_replication_conf if $params{hba_permit_replication};
+	$self->set_replication_conf if $params{allows_streaming};
 	$self->enable_archiving     if $params{has_archiving};
 }
 
@@ -486,7 +485,7 @@ sub append_conf
 =item $node->backup(backup_name)
 
 Create a hot backup with B<pg_basebackup> in subdirectory B<backup_name> of
-B<< $node->backup_dir >>, including the transaction logs. Transaction logs are
+B<< $node->backup_dir >>, including the WAL. WAL files
 fetched at the end of the backup, not streamed.
 
 You'll have to configure a suitable B<max_wal_senders> on the
@@ -498,18 +497,19 @@ sub backup
 {
 	my ($self, $backup_name) = @_;
 	my $backup_path = $self->backup_dir . '/' . $backup_name;
+	my $port        = $self->port;
 	my $name        = $self->name;
 
 	print "# Taking pg_basebackup $backup_name from node \"$name\"\n";
-	TestLib::system_or_bail('pg_basebackup', '-D', $backup_path, '-h',
-		$self->host, '-p', $self->port, '-x');
+	TestLib::system_or_bail('pg_basebackup', '-D', $backup_path, '-p', $port,
+		'--no-sync');
 	print "# Backup finished\n";
 }
 
 =item $node->backup_fs_hot(backup_name)
 
 Create a backup with a filesystem level copy in subdirectory B<backup_name> of
-B<< $node->backup_dir >>, including transaction logs.
+B<< $node->backup_dir >>, including WAL.
 
 Archiving must be enabled, as B<pg_start_backup()> and B<pg_stop_backup()> are
 used. This is not checked or enforced.
@@ -527,7 +527,7 @@ sub backup_fs_hot
 =item $node->backup_fs_cold(backup_name)
 
 Create a backup with a filesystem level copy in subdirectory B<backup_name> of
-B<< $node->backup_dir >>, including transaction logs. The server must be
+B<< $node->backup_dir >>, including WAL. The server must be
 stopped as no attempt to handle concurrent writes is made.
 
 Use B<backup> or B<backup_fs_hot> if you want to back up a running server.
@@ -563,7 +563,7 @@ sub _backup_fs
 		$backup_path,
 		filterfn => sub {
 			my $src = shift;
-			return ($src ne 'pg_log' and $src ne 'postmaster.pid');
+			return ($src ne 'log' and $src ne 'postmaster.pid');
 		});
 
 	if ($hot)
@@ -594,9 +594,6 @@ Does not start the node after initializing it.
 
 A recovery.conf is not created.
 
-pg_hba.conf is configured to allow replication connections. Pass the keyword
-parameter hba_permit_replication => 0 to disable this.
-
 Streaming replication can be enabled on this node by passing the keyword
 parameter has_streaming => 1. This is disabled by default.
 
@@ -613,14 +610,11 @@ sub init_from_backup
 {
 	my ($self, $root_node, $backup_name, %params) = @_;
 	my $backup_path = $root_node->backup_dir . '/' . $backup_name;
-	my $host        = $self->host;
 	my $port        = $self->port;
 	my $node_name   = $self->name;
 	my $root_name   = $root_node->name;
 
 	$params{has_streaming} = 0 unless defined $params{has_streaming};
-	$params{hba_permit_replication} = 1
-	  unless defined $params{hba_permit_replication};
 	$params{has_restoring} = 0 unless defined $params{has_restoring};
 
 	print
@@ -642,102 +636,39 @@ sub init_from_backup
 		qq(
 port = $port
 ));
-	if ($use_tcp)
-	{
-		$self->append_conf('postgresql.conf', "listen_addresses = '$host'");
-	}
-	else
-	{
-		$self->append_conf('postgresql.conf',
-			"unix_socket_directories = '$host'");
-	}
-	$self->set_replication_conf         if $params{hba_permit_replication};
 	$self->enable_streaming($root_node) if $params{has_streaming};
 	$self->enable_restoring($root_node) if $params{has_restoring};
 }
 
 =pod
 
-=item $node->rotate_logfile()
+=item $node->start()
 
-Switch to a new PostgreSQL log file.  This does not alter any running
-PostgreSQL process.  Subsequent method calls, including pg_ctl invocations,
-will use the new name.  Return the new name.
-
-=cut
-
-sub rotate_logfile
-{
-	my ($self) = @_;
-	$self->{_logfile} = sprintf('%s_%d.log',
-		$self->{_logfile_base},
-		++$self->{_logfile_generation});
-	return $self->{_logfile};
-}
-
-=pod
-
-=item $node->start(%params) => success_or_failure
-
-Wrapper for pg_ctl -w start
+Wrapper for pg_ctl start
 
 Start the node and wait until it is ready to accept connections.
 
-=over
-
-=item fail_ok => 1
-
-By default, failure terminates the entire F<prove> invocation.  If given,
-instead return a true or false value to indicate success or failure.
-
-=back
-
 =cut
 
 sub start
 {
-	my ($self, %params) = @_;
+	my ($self) = @_;
 	my $port   = $self->port;
 	my $pgdata = $self->data_dir;
 	my $name   = $self->name;
 	BAIL_OUT("node \"$name\" is already running") if defined $self->{_pid};
 	print("### Starting node \"$name\"\n");
-	my $ret = TestLib::system_log('pg_ctl', '-w', '-D', $self->data_dir, '-l',
+	my $ret = TestLib::system_log('pg_ctl', '-D', $self->data_dir, '-l',
 		$self->logfile, 'start');
 
 	if ($ret != 0)
 	{
 		print "# pg_ctl start failed; logfile:\n";
 		print TestLib::slurp_file($self->logfile);
-		BAIL_OUT("pg_ctl start failed") unless $params{fail_ok};
-		return 0;
+		BAIL_OUT("pg_ctl start failed");
 	}
 
 	$self->_update_pid(1);
-	return 1;
-}
-
-=pod
-
-=item $node->kill9()
-
-Send SIGKILL (signal 9) to the postmaster.
-
-Note: if the node is already known stopped, this does nothing.
-However, if we think it's running and it's not, it's important for
-this to fail.  Otherwise, tests might fail to detect server crashes.
-
-=cut
-
-sub kill9
-{
-	my ($self) = @_;
-	my $name = $self->name;
-	return unless defined $self->{_pid};
-	print "### Killing node \"$name\" using signal 9\n";
-	kill(9, $self->{_pid}) or BAIL_OUT("kill(9, $self->{_pid}) failed");
-	$self->{_pid} = undef;
-	return;
 }
 
 =pod
@@ -787,7 +718,7 @@ sub reload
 
 =item $node->restart()
 
-Wrapper for pg_ctl -w restart
+Wrapper for pg_ctl restart
 
 =cut
 
@@ -799,8 +730,8 @@ sub restart
 	my $logfile = $self->logfile;
 	my $name    = $self->name;
 	print "### Restarting node \"$name\"\n";
-	TestLib::system_or_bail('pg_ctl', '-D', $pgdata, '-w', '-l', $logfile,
-							'restart');
+	TestLib::system_or_bail('pg_ctl', '-D', $pgdata, '-l', $logfile,
+		'restart');
 	$self->_update_pid(1);
 }
 
@@ -821,7 +752,7 @@ sub promote
 	my $name    = $self->name;
 	print "### Promoting node \"$name\"\n";
 	TestLib::system_or_bail('pg_ctl', '-D', $pgdata, '-l', $logfile,
-							'promote');
+		'promote');
 }
 
 # Internal routine to enable streaming replication on a standby node.
@@ -843,7 +774,7 @@ standby_mode=on
 sub enable_restoring
 {
 	my ($self, $root_node) = @_;
-	my $path = TestLib::perl2host($root_node->archive_dir);
+	my $path = $vfs_path . $root_node->archive_dir;
 	my $name = $self->name;
 
 	print "### Enabling WAL restore for node \"$name\"\n";
@@ -871,7 +802,7 @@ standby_mode = on
 sub enable_archiving
 {
 	my ($self) = @_;
-	my $path   = TestLib::perl2host($self->archive_dir);
+	my $path   = $vfs_path . $self->archive_dir;
 	my $name   = $self->name;
 
 	print "### Enabling WAL archiving for node \"$name\"\n";
@@ -917,13 +848,14 @@ sub _update_pid
 
 	$self->{_pid} = undef;
 	print "# No postmaster PID for node \"$name\"\n";
+
 	# Complain if we expected to find a pidfile.
 	BAIL_OUT("postmaster.pid unexpectedly not present") if $is_running;
 }
 
 =pod
 
-=item PostgresNode->get_new_node(node_name, %params)
+=item PostgresNode->get_new_node(node_name)
 
 Build a new object of class C<PostgresNode> (or of a subclass, if you have
 one), assigning a free port number.  Remembers the node, to prevent its port
@@ -932,22 +864,6 @@ shut down when the test script exits.
 
 You should generally use this instead of C<PostgresNode::new(...)>.
 
-=over
-
-=item port => [1,65535]
-
-By default, this function assigns a port number to each node.  Specify this to
-force a particular port number.  The caller is responsible for evaluating
-potential conflicts and privilege requirements.
-
-=item own_host => 1
-
-By default, all nodes use the same PGHOST value.  If specified, generate a
-PGHOST specific to this node.  This allows multiple nodes to use the same
-port.
-
-=back
-
 For backwards compatibility, it is also exported as a standalone function,
 which can only create objects of class C<PostgresNode>.
 
@@ -956,11 +872,10 @@ which can only create objects of class C<PostgresNode>.
 sub get_new_node
 {
 	my $class = 'PostgresNode';
-	$class = shift if scalar(@_) % 2 != 1;
-	my ($name, %params) = @_;
-	my $port_is_forced = defined $params{port};
-	my $found          = $port_is_forced;
-	my $port = $port_is_forced ? $params{port} : $last_port_assigned;
+	$class = shift if 1 < scalar @_;
+	my $name  = shift;
+	my $found = 0;
+	my $port  = $last_port_assigned;
 
 	while ($found == 0)
 	{
@@ -978,79 +893,41 @@ sub get_new_node
 		}
 
 		# Check to see if anything else is listening on this TCP port.
-		# Seek a port available for all possible listen_addresses values,
-		# so callers can harness this port for the widest range of purposes.
-		# The 0.0.0.0 test achieves that for MSYS, which automatically sets
-		# SO_EXCLUSIVEADDRUSE.  Testing 0.0.0.0 is insufficient for Windows
-		# native Perl (https://stackoverflow.com/a/14388707), so we also
-		# have to test individual addresses.  Doing that for 127.0.0/24
-		# addresses other than 127.0.0.1 might fail with EADDRNOTAVAIL on
-		# non-Linux, non-Windows kernels.
-		#
-		# Thus, 0.0.0.0 and individual 127.0.0/24 addresses are tested
-		# only on Windows and only when TCP usage is requested.
+		# This is *necessary* on Windows, and seems like a good idea
+		# on Unixen as well, even though we don't ask the postmaster
+		# to open a TCP port on Unix.
 		if ($found == 1)
 		{
-			foreach my $addr (qw(127.0.0.1),
-               ($use_tcp && $TestLib::windows_os)
-               ? qw(127.0.0.2 127.0.0.3 0.0.0.0)
-               : ())
-			{
-				can_bind($addr, $port) or $found = 0;
-			}
+			my $iaddr = inet_aton($test_localhost);
+			my $paddr = sockaddr_in($port, $iaddr);
+			my $proto = getprotobyname("tcp");
+
+			socket(SOCK, PF_INET, SOCK_STREAM, $proto)
+			  or die "socket failed: $!";
+
+			# As in postmaster, don't use SO_REUSEADDR on Windows
+			setsockopt(SOCK, SOL_SOCKET, SO_REUSEADDR, pack("l", 1))
+			  unless $TestLib::windows_os;
+			(bind(SOCK, $paddr) && listen(SOCK, SOMAXCONN))
+			  or $found = 0;
+			close(SOCK);
 		}
 	}
 
-	print "# Found port $port\n";
-
-	# Select a host.
-	my $host = $test_pghost;
-	if ($params{own_host})
-	{
-		if ($use_tcp)
-		{
-			$last_host_assigned++;
-			$last_host_assigned > 254 and BAIL_OUT("too many own_host nodes");
-			$host = '127.0.0.' . $last_host_assigned;
-		}
-		else
-		{
-			$host = "$test_pghost/$name"; # Assume $name =~ /^[-_a-zA-Z0-9]+$/
-			mkdir $host;
-		}
-	}
+	print "# Found free port $port\n";
 
 	# Lock port number found by creating a new node
-	my $node = $class->new($name, $host, $port);
+	my $node = $class->new($name, $test_pghost, $port);
 
 	# Add node to list of nodes
 	push(@all_nodes, $node);
 
 	# And update port for next time
-	$port_is_forced or $last_port_assigned = $port;
+	$last_port_assigned = $port;
 
 	return $node;
 }
 
-# Internal routine to check whether a host:port is available to bind
-sub can_bind
-{
-	my ($host, $port) = @_;
-	my $iaddr = inet_aton($host);
-	my $paddr = sockaddr_in($port, $iaddr);
-	my $proto = getprotobyname("tcp");
-
-	socket(SOCK, PF_INET, SOCK_STREAM, $proto)
-	  or die "socket failed: $!";
-
-	# As in postmaster, don't use SO_REUSEADDR on Windows
-	setsockopt(SOCK, SOL_SOCKET, SO_REUSEADDR, pack("l", 1))
-	  unless $TestLib::windows_os;
-	my $ret = bind(SOCK, $paddr) && listen(SOCK, SOMAXCONN);
-	close(SOCK);
-	return $ret;
-}
-
 # Automatically shut down any still-running nodes when the test script exits.
 # Note that this just stops the postmasters (in the same order the nodes were
 # created in).  Temporary PGDATA directories are deleted, in an unspecified
@@ -1118,6 +995,7 @@ sub safe_psql
 		print "\n#### End standard error\n";
 	}
 
+	$stdout =~ s/\r//g if $TestLib::windows_os;
 	return $stdout;
 }
 
@@ -1273,7 +1151,8 @@ sub psql
 			# IPC::Run::run threw an exception. re-throw unless it's a
 			# timeout, which we'll handle by testing is_expired
 			die $exc_save
-			  if (blessed($exc_save) || $exc_save !~ /^\Q$timeout_exception\E/);
+			  if (blessed($exc_save)
+				|| $exc_save !~ /^\Q$timeout_exception\E/);
 
 			$ret = undef;
 
@@ -1292,20 +1171,16 @@ sub psql
 		}
 	};
 
-	# Note: on Windows, IPC::Run seems to convert \r\n to \n in program output
-	# if we're using native Perl, but not if we're using MSys Perl.  So do it
-	# by hand in the latter case, here and elsewhere.
-
 	if (defined $$stdout)
 	{
-		$$stdout =~ s/\r\n/\n/g if $Config{osname} eq 'msys';
 		chomp $$stdout;
+		$$stdout =~ s/\r//g if $TestLib::windows_os;
 	}
 
 	if (defined $$stderr)
 	{
-		$$stderr =~ s/\r\n/\n/g if $Config{osname} eq 'msys';
 		chomp $$stderr;
+		$$stderr =~ s/\r//g if $TestLib::windows_os;
 	}
 
 	# See http://perldoc.perl.org/perlvar.html#%24CHILD_ERROR
@@ -1325,7 +1200,8 @@ sub psql
 		  if $ret == 1;
 		die "connection error: '$$stderr'\nwhile running '@psql_params'"
 		  if $ret == 2;
-		die "error running SQL: '$$stderr'\nwhile running '@psql_params'"
+		die
+"error running SQL: '$$stderr'\nwhile running '@psql_params' with sql '$sql'"
 		  if $ret == 3;
 		die "psql returns $ret: '$$stderr'\nwhile running '@psql_params'";
 	}
@@ -1342,277 +1218,48 @@ sub psql
 
 =pod
 
-=item $node->background_psql($dbname, \$stdin, \$stdout, $timer, %params) => harness
+=item $node->poll_query_until($dbname, $query [, $expected ])
 
-Invoke B<psql> on B<$dbname> and return an IPC::Run harness object, which the
-caller may use to send input to B<psql>.  The process's stdin is sourced from
-the $stdin scalar reference, and its stdout and stderr go to the $stdout
-scalar reference.  This allows the caller to act on other parts of the system
-while idling this backend.
-
-The specified timer object is attached to the harness, as well.  It's caller's
-responsibility to select the timeout length, and to restart the timer after
-each command if the timeout is per-command.
-
-psql is invoked in tuples-only unaligned mode with reading of B<.psqlrc>
-disabled.  That may be overridden by passing extra psql parameters.
-
-Dies on failure to invoke psql, or if psql fails to connect.  Errors occurring
-later are the caller's problem.  psql runs with on_error_stop by default so
-that it will stop running sql and return 3 if passed SQL results in an error.
-
-Be sure to "finish" the harness when done with it.
-
-=over
-
-=item on_error_stop => 1
-
-By default, the B<psql> method invokes the B<psql> program with ON_ERROR_STOP=1
-set, so SQL execution is stopped at the first error and exit code 3 is
-returned.  Set B<on_error_stop> to 0 to ignore errors instead.
-
-=item replication => B<value>
-
-If set, add B<replication=value> to the conninfo string.
-Passing the literal value C<database> results in a logical replication
-connection.
-
-=item extra_params => ['--single-transaction']
-
-If given, it must be an array reference containing additional parameters to B<psql>.
-
-=back
-
-=cut
-
-sub background_psql
-{
-	my ($self, $dbname, $stdin, $stdout, $timer, %params) = @_;
-
-	local $ENV{PGHOST} = $self->host;
-	local $ENV{PGPORT} = $self->port;
-
-	my $replication = $params{replication};
-
-	my @psql_params = (
-		'psql',
-		'-XAtq',
-		'-d',
-		$self->connstr($dbname)
-		  . (defined $replication ? " replication=$replication" : ""),
-		'-f',
-		'-');
-
-	$params{on_error_stop} = 1 unless defined $params{on_error_stop};
-
-	push @psql_params, '-v', 'ON_ERROR_STOP=1' if $params{on_error_stop};
-	push @psql_params, @{ $params{extra_params} }
-	  if defined $params{extra_params};
-
-	# Ensure there is no data waiting to be sent:
-	$$stdin = "" if ref($stdin);
-	# IPC::Run would otherwise append to existing contents:
-	$$stdout = "" if ref($stdout);
-
-	my $harness = IPC::Run::start \@psql_params,
-	  '<', $stdin, '>', $stdout, $timer;
-
-	# Request some output, and pump until we see it.  This means that psql
-	# connection failures are caught here, relieving callers of the need to
-	# handle those.  (Right now, we have no particularly good handling for
-	# errors anyway, but that might be added later.)
-	my $banner = "background_psql: ready";
-	$$stdin = "\\echo $banner\n";
-	pump $harness until $$stdout =~ /$banner/ || $timer->is_expired;
-
-	die "psql startup timed out" if $timer->is_expired;
-
-	return $harness;
-}
-
-# Common sub of pgbench-invoking interfaces.  Makes any requested script files
-# and returns pgbench command-line options causing use of those files.
-sub _pgbench_make_files
-{
-	my ($self, $files) = @_;
-	my @file_opts;
-
-	if (defined $files)
-	{
-
-		# note: files are ordered for determinism
-		for my $fn (sort keys %$files)
-		{
-			my $filename = $self->basedir . '/' . $fn;
-			push @file_opts, '-f', $filename;
-
-			# cleanup file weight
-			$filename =~ s/\@\d+$//;
-
-			#push @filenames, $filename;
-			# filenames are expected to be unique on a test
-			if (-e $filename)
-			{
-				ok(0, "$filename must not already exist");
-				unlink $filename or die "cannot unlink $filename: $!";
-			}
-			TestLib::append_to_file($filename, $$files{$fn});
-		}
-	}
-
-	return @file_opts;
-}
-
-=pod
-
-=item $node->pgbench($opts, $stat, $out, $err, $name, $files, @args)
-
-Invoke B<pgbench>, with parameters and files.
-
-=over
-
-=item $opts
-
-Options as a string to be split on spaces.
-
-=item $stat
-
-Expected exit status.
-
-=item $out
-
-Reference to a regexp list that must match stdout.
-
-=item $err
-
-Reference to a regexp list that must match stderr.
-
-=item $name
-
-Name of test for error messages.
-
-=item $files
-
-Reference to filename/contents dictionary.
-
-=item @args
-
-Further raw options or arguments.
-
-=back
-
-=cut
-
-sub pgbench
-{
-	local $Test::Builder::Level = $Test::Builder::Level + 1;
-
-	my ($self, $opts, $stat, $out, $err, $name, $files, @args) = @_;
-	my @cmd = (
-		'pgbench',
-		split(/\s+/, $opts),
-		$self->_pgbench_make_files($files), @args);
-
-	$self->command_checks_all(\@cmd, $stat, $out, $err, $name);
-}
-
-=pod
-
-=item $node->background_pgbench($opts, $files, \$stdout, $timer) => harness
-
-Invoke B<pgbench> and return an IPC::Run harness object.  The process's stdin
-is empty, and its stdout and stderr go to the $stdout scalar reference.  This
-allows the caller to act on other parts of the system while B<pgbench> is
-running.  Errors from B<pgbench> are the caller's problem.
-
-The specified timer object is attached to the harness, as well.  It's caller's
-responsibility to select the timeout length, and to restart the timer after
-each command if the timeout is per-command.
-
-Be sure to "finish" the harness when done with it.
-
-=over
-
-=item $opts
-
-Options as a string to be split on spaces.
-
-=item $files
-
-Reference to filename/contents dictionary.
-
-=back
-
-=cut
-
-sub background_pgbench
-{
-	my ($self, $opts, $files, $stdout, $timer) = @_;
-
-	my @cmd =
-	  ('pgbench', split(/\s+/, $opts), $self->_pgbench_make_files($files));
-
-	local $ENV{PGHOST} = $self->host;
-	local $ENV{PGPORT} = $self->port;
-
-	my $stdin = "";
-	# IPC::Run would otherwise append to existing contents:
-	$$stdout = "" if ref($stdout);
-
-	my $harness = IPC::Run::start \@cmd, '<', \$stdin, '>', $stdout, '2>&1',
-	  $timer;
-
-	return $harness;
-}
-
-=pod
-
-=item $node->poll_query_until(dbname, query)
-
-Run a query once a second, until it returns 't' (i.e. SQL boolean true).
-Continues polling if psql returns an error result. Times out after 180 seconds.
+Run B<$query> repeatedly, until it returns the B<$expected> result
+('t', or SQL boolean true, by default).
+Continues polling if B<psql> returns an error result.
+Times out after 180 seconds.
+Returns 1 if successful, 0 if timed out.
 
 =cut
 
 sub poll_query_until
 {
-	my ($self, $dbname, $query) = @_;
+	my ($self, $dbname, $query, $expected) = @_;
 
-	my $max_attempts = 180;
-	my $attempts     = 0;
+	$expected = 't' unless defined($expected);    # default value
+
+	my $cmd = [ 'psql', '-XAt', '-c', $query, '-d', $self->connstr($dbname) ];
 	my ($stdout, $stderr);
+	my $max_attempts = 180 * 10;
+	my $attempts     = 0;
 
 	while ($attempts < $max_attempts)
 	{
-		my $cmd =
-		  [ 'psql', '-XAt', '-d', $self->connstr($dbname) ];
-		my $result = IPC::Run::run $cmd, '<', \$query,
-		  '>', \$stdout, '2>', \$stderr;
+		my $result = IPC::Run::run $cmd, '>', \$stdout, '2>', \$stderr;
 
-		$stdout =~ s/\r\n/\n/g if $Config{osname} eq 'msys';
 		chomp($stdout);
-		if ($stdout eq "t")
+		$stdout =~ s/\r//g if $TestLib::windows_os;
+
+		if ($stdout eq $expected)
 		{
 			return 1;
 		}
 
-		# Wait a second before retrying.
-		sleep 1;
+		# Wait 0.1 second before retrying.
+		usleep(100_000);
+
 		$attempts++;
 	}
 
-	# The query result didn't change in 180 seconds. Give up. Print the
-	# output from the last attempt, hopefully that's useful for debugging.
-	$stderr =~ s/\r\n/\n/g if $Config{osname} eq 'msys';
-	chomp($stderr);
-	diag qq(poll_query_until timed out executing this query:
-$query
-expecting this output:
-t
-last actual query output:
-$stdout
-with stderr:
-$stderr);
+	# The query result didn't change in 180 seconds. Give up. Print the stderr
+	# from the last attempt, hopefully that's useful for debugging.
+	diag $stderr;
 	return 0;
 }
 
@@ -1620,8 +1267,9 @@ $stderr);
 
 =item $node->command_ok(...)
 
-Runs a shell command like TestLib::command_ok, but with PGHOST and PGPORT set
-so that the command will default to connecting to this PostgresNode.
+Runs a shell command like TestLib::command_ok, but with PGPORT
+set so that the command will default to connecting to this
+PostgresNode.
 
 =cut
 
@@ -1629,7 +1277,6 @@ sub command_ok
 {
 	my $self = shift;
 
-	local $ENV{PGHOST} = $self->host;
 	local $ENV{PGPORT} = $self->port;
 
 	TestLib::command_ok(@_);
@@ -1639,7 +1286,7 @@ sub command_ok
 
 =item $node->command_fails(...) - TestLib::command_fails with our PGPORT
 
-TestLib::command_fails with our connection parameters. See command_ok(...)
+See command_ok(...)
 
 =cut
 
@@ -1647,7 +1294,6 @@ sub command_fails
 {
 	my $self = shift;
 
-	local $ENV{PGHOST} = $self->host;
 	local $ENV{PGPORT} = $self->port;
 
 	TestLib::command_fails(@_);
@@ -1657,7 +1303,7 @@ sub command_fails
 
 =item $node->command_like(...)
 
-TestLib::command_like with our connection parameters. See command_ok(...)
+TestLib::command_like with our PGPORT. See command_ok(...)
 
 =cut
 
@@ -1665,7 +1311,6 @@ sub command_like
 {
 	my $self = shift;
 
-	local $ENV{PGHOST} = $self->host;
 	local $ENV{PGPORT} = $self->port;
 
 	TestLib::command_like(@_);
@@ -1673,50 +1318,49 @@ sub command_like
 
 =pod
 
-=item $node->command_checks_all(...)
-
-TestLib::command_checks_all with our connection parameters. See
-command_ok(...)
-
-=cut
-
-sub command_checks_all
-{
-	my $self = shift;
-
-	local $ENV{PGHOST} = $self->host;
-	local $ENV{PGPORT} = $self->port;
-
-	TestLib::command_checks_all(@_);
-	return;
-}
-
-=pod
-
 =item $node->issues_sql_like(cmd, expected_sql, test_name)
 
 Run a command on the node, then verify that $expected_sql appears in the
 server log file.
 
+Reads the whole log file so be careful when working with large log outputs.
+The log file is truncated prior to running the command, however.
+
 =cut
 
 sub issues_sql_like
 {
 	my ($self, $cmd, $expected_sql, $test_name) = @_;
 
-	local $ENV{PGHOST} = $self->host;
 	local $ENV{PGPORT} = $self->port;
 
-	my $log_location = -s $self->logfile;
-
+	truncate $self->logfile, 0;
 	my $result = TestLib::run_log($cmd);
 	ok($result, "@$cmd exit code 0");
-	my $log = TestLib::slurp_file($self->logfile, $log_location);
+	my $log = TestLib::slurp_file($self->logfile);
 	like($log, $expected_sql, "$test_name: SQL found in server log");
 }
 
 =pod
 
+=item $node->run_log(...)
+
+Runs a shell command like TestLib::run_log, but with PGPORT set so
+that the command will default to connecting to this PostgresNode.
+
+=cut
+
+sub run_log
+{
+	my $self = shift;
+
+	local $ENV{PGPORT} = $self->port;
+
+	TestLib::run_log(@_);
+}
+
+=pod
+
 =item $node->lsn(mode)
 
 Look up WAL locations on the server:
@@ -1735,11 +1379,11 @@ sub lsn
 {
 	my ($self, $mode) = @_;
 	my %modes = (
-		'insert'  => 'pg_current_xlog_insert_location()',
-		'flush'   => 'pg_current_xlog_flush_location()',
-		'write'   => 'pg_current_xlog_location()',
-		'receive' => 'pg_last_xlog_receive_location()',
-		'replay'  => 'pg_last_xlog_replay_location()');
+		'insert'  => 'pg_current_wal_insert_lsn()',
+		'flush'   => 'pg_current_wal_flush_lsn()',
+		'write'   => 'pg_current_wal_lsn()',
+		'receive' => 'pg_last_wal_receive_lsn()',
+		'replay'  => 'pg_last_wal_replay_lsn()');
 
 	$mode = '<undef>' if !defined($mode);
 	die "unknown mode for 'lsn': '$mode', valid modes are "
@@ -1763,10 +1407,10 @@ sub lsn
 =item $node->wait_for_catchup(standby_name, mode, target_lsn)
 
 Wait for the node with application_name standby_name (usually from node->name)
-until its replication position in pg_stat_replication equals or passes the
-upstream's xlog insert point at the time this function is called. By default
-the replay_location is waited for, but 'mode' may be specified to wait for any
-of sent|write|flush|replay.
+until its replication location in pg_stat_replication equals or passes the
+upstream's WAL insert point at the time this function is called. By default
+the replay_lsn is waited for, but 'mode' may be specified to wait for any of
+sent|write|flush|replay.
 
 If there is no active replication connection from this peer, waits until
 poll_query_until timeout.
@@ -1783,23 +1427,242 @@ sub wait_for_catchup
 {
 	my ($self, $standby_name, $mode, $target_lsn) = @_;
 	$mode = defined($mode) ? $mode : 'replay';
-	my %valid_modes = ( 'sent' => 1, 'write' => 1, 'flush' => 1, 'replay' => 1 );
-	die "unknown mode $mode for 'wait_for_catchup', valid modes are " . join(', ', keys(%valid_modes)) unless exists($valid_modes{$mode});
+	my %valid_modes =
+	  ('sent' => 1, 'write' => 1, 'flush' => 1, 'replay' => 1);
+	die "unknown mode $mode for 'wait_for_catchup', valid modes are "
+	  . join(', ', keys(%valid_modes))
+	  unless exists($valid_modes{$mode});
+
 	# Allow passing of a PostgresNode instance as shorthand
-	if ( blessed( $standby_name ) && $standby_name->isa("PostgresNode") )
+	if (blessed($standby_name) && $standby_name->isa("PostgresNode"))
 	{
 		$standby_name = $standby_name->name;
 	}
 	die 'target_lsn must be specified' unless defined($target_lsn);
-	print "Waiting for replication conn " . $standby_name . "'s " . $mode . "_location to pass " . $target_lsn . " on " . $self->name . "\n";
-	my $query = qq[SELECT '$target_lsn' <= ${mode}_location FROM pg_catalog.pg_stat_replication WHERE application_name = '$standby_name';];
+	print "Waiting for replication conn "
+	  . $standby_name . "'s "
+	  . $mode
+	  . "_lsn to pass "
+	  . $target_lsn . " on "
+	  . $self->name . "\n";
+	my $query =
+qq[SELECT '$target_lsn' <= ${mode}_lsn FROM pg_catalog.pg_stat_replication WHERE application_name = '$standby_name';];
 	$self->poll_query_until('postgres', $query)
-		or die "timed out waiting for catchup, current position is " . ($self->safe_psql('postgres', $query) || '(unknown)');
+	  or die "timed out waiting for catchup, current location is "
+	  . ($self->safe_psql('postgres', $query) || '(unknown)');
 	print "done\n";
 }
 
 =pod
 
+=item $node->wait_for_slot_catchup(slot_name, mode, target_lsn)
+
+Wait for the named replication slot to equal or pass the supplied target_lsn.
+The location used is the restart_lsn unless mode is given, in which case it may
+be 'restart' or 'confirmed_flush'.
+
+Requires that the 'postgres' db exists and is accessible.
+
+This is not a test. It die()s on failure.
+
+If the slot is not active, will time out after poll_query_until's timeout.
+
+target_lsn may be any arbitrary lsn, but is typically $master_node->lsn('insert').
+
+Note that for logical slots, restart_lsn is held down by the oldest in-progress tx.
+
+=cut
+
+sub wait_for_slot_catchup
+{
+	my ($self, $slot_name, $mode, $target_lsn) = @_;
+	$mode = defined($mode) ? $mode : 'restart';
+	if (!($mode eq 'restart' || $mode eq 'confirmed_flush'))
+	{
+		die "valid modes are restart, confirmed_flush";
+	}
+	die 'target lsn must be specified' unless defined($target_lsn);
+	print "Waiting for replication slot "
+	  . $slot_name . "'s "
+	  . $mode
+	  . "_lsn to pass "
+	  . $target_lsn . " on "
+	  . $self->name . "\n";
+	my $query =
+qq[SELECT '$target_lsn' <= ${mode}_lsn FROM pg_catalog.pg_replication_slots WHERE slot_name = '$slot_name';];
+	$self->poll_query_until('postgres', $query)
+	  or die "timed out waiting for catchup, current location is "
+	  . ($self->safe_psql('postgres', $query) || '(unknown)');
+	print "done\n";
+}
+
+=pod
+
+=item $node->query_hash($dbname, $query, @columns)
+
+Execute $query on $dbname, replacing any appearance of the string __COLUMNS__
+within the query with a comma-separated list of @columns.
+
+If __COLUMNS__ does not appear in the query, its result columns must EXACTLY
+match the order and number (but not necessarily alias) of supplied @columns.
+
+The query must return zero or one rows.
+
+Return a hash-ref representation of the results of the query, with any empty
+or null results as defined keys with an empty-string value. There is no way
+to differentiate between null and empty-string result fields.
+
+If the query returns zero rows, return a hash with all columns empty. There
+is no way to differentiate between zero rows returned and a row with only
+null columns.
+
+=cut
+
+sub query_hash
+{
+	my ($self, $dbname, $query, @columns) = @_;
+	die 'calls in array context for multi-row results not supported yet'
+	  if (wantarray);
+
+	# Replace __COLUMNS__ if found
+	substr($query, index($query, '__COLUMNS__'), length('__COLUMNS__')) =
+	  join(', ', @columns)
+	  if index($query, '__COLUMNS__') >= 0;
+	my $result = $self->safe_psql($dbname, $query);
+
+	# hash slice, see http://stackoverflow.com/a/16755894/398670 .
+	#
+	# Fills the hash with empty strings produced by x-operator element
+	# duplication if result is an empty row
+	#
+	my %val;
+	@val{@columns} =
+	  $result ne '' ? split(qr/\|/, $result, -1) : ('',) x scalar(@columns);
+	return \%val;
+}
+
+=pod
+
+=item $node->slot(slot_name)
+
+Return hash-ref of replication slot data for the named slot, or a hash-ref with
+all values '' if not found. Does not differentiate between null and empty string
+for fields, no field is ever undef.
+
+The restart_lsn and confirmed_flush_lsn fields are returned verbatim, and also
+as a 2-list of [highword, lowword] integer. Since we rely on Perl 5.8.8 we can't
+"use bigint", it's from 5.20, and we can't assume we have Math::Bigint from CPAN
+either.
+
+=cut
+
+sub slot
+{
+	my ($self, $slot_name) = @_;
+	my @columns = (
+		'plugin', 'slot_type',  'datoid', 'database',
+		'active', 'active_pid', 'xmin',   'catalog_xmin',
+		'restart_lsn');
+	return $self->query_hash(
+		'postgres',
+"SELECT __COLUMNS__ FROM pg_catalog.pg_replication_slots WHERE slot_name = '$slot_name'",
+		@columns);
+}
+
+=pod
+
+=item $node->pg_recvlogical_upto(self, dbname, slot_name, endpos, timeout_secs, ...)
+
+Invoke pg_recvlogical to read from slot_name on dbname until LSN endpos, which
+corresponds to pg_recvlogical --endpos.  Gives up after timeout (if nonzero).
+
+Disallows pg_recvlogical from internally retrying on error by passing --no-loop.
+
+Plugin options are passed as additional keyword arguments.
+
+If called in scalar context, returns stdout, and die()s on timeout or nonzero return.
+
+If called in array context, returns a tuple of (retval, stdout, stderr, timeout).
+timeout is the IPC::Run::Timeout object whose is_expired method can be tested
+to check for timeout. retval is undef on timeout.
+
+=cut
+
+sub pg_recvlogical_upto
+{
+	my ($self, $dbname, $slot_name, $endpos, $timeout_secs, %plugin_options) =
+	  @_;
+	my ($stdout, $stderr);
+
+	my $timeout_exception = 'pg_recvlogical timed out';
+
+	die 'slot name must be specified' unless defined($slot_name);
+	die 'endpos must be specified'    unless defined($endpos);
+
+	my @cmd = (
+		'pg_recvlogical', '-S', $slot_name, '--dbname',
+		$self->connstr($dbname));
+	push @cmd, '--endpos', $endpos;
+	push @cmd, '-f', '-', '--no-loop', '--start';
+
+	while (my ($k, $v) = each %plugin_options)
+	{
+		die "= is not permitted to appear in replication option name"
+		  if ($k =~ qr/=/);
+		push @cmd, "-o", "$k=$v";
+	}
+
+	my $timeout;
+	$timeout =
+	  IPC::Run::timeout($timeout_secs, exception => $timeout_exception)
+	  if $timeout_secs;
+	my $ret = 0;
+
+	do
+	{
+		local $@;
+		eval {
+			IPC::Run::run(\@cmd, ">", \$stdout, "2>", \$stderr, $timeout);
+			$ret = $?;
+		};
+		my $exc_save = $@;
+		if ($exc_save)
+		{
+
+			# IPC::Run::run threw an exception. re-throw unless it's a
+			# timeout, which we'll handle by testing is_expired
+			die $exc_save
+			  if (blessed($exc_save) || $exc_save !~ qr/$timeout_exception/);
+
+			$ret = undef;
+
+			die "Got timeout exception '$exc_save' but timer not expired?!"
+			  unless $timeout->is_expired;
+
+			die
+"$exc_save waiting for endpos $endpos with stdout '$stdout', stderr '$stderr'"
+			  unless wantarray;
+		}
+	};
+
+	$stdout =~ s/\r//g if $TestLib::windows_os;
+	$stderr =~ s/\r//g if $TestLib::windows_os;
+
+	if (wantarray)
+	{
+		return ($ret, $stdout, $stderr, $timeout);
+	}
+	else
+	{
+		die
+"pg_recvlogical exited with code '$ret', stdout '$stdout' and stderr '$stderr'"
+		  if $ret;
+		return $stdout;
+	}
+}
+
+=pod
+
 =back
 
 =cut
