@@ -116,14 +116,7 @@ insert into bmscantest select r, 'fooooooooooooooooooooooooooooooooooooooooooooo
 create index i_bmtest ON bmscantest(a);
 select count(*) from bmscantest where a>1;
 
--- test accumulation of stats for parallel nodes
 reset enable_seqscan;
-alter table tenk2 set (parallel_workers = 0);
-explain (analyze, timing off, summary off, costs off)
-   select count(*) from tenk1, tenk2 where tenk1.hundred > 1
-        and tenk2.thousand=0;
-alter table tenk2 reset (parallel_workers);
-
 reset enable_indexscan;
 reset enable_hashjoin;
 reset enable_mergejoin;
@@ -151,13 +144,6 @@ explain (costs off)
 
 select count(*) from tenk1 group by twenty;
 
--- test handling of SRFs in targetlist (bug in 10.0)
-
-explain (costs off)
-   select count(*), generate_series(1,2) from tenk1 group by twenty;
-
-select count(*), generate_series(1,2) from tenk1 group by twenty;
-
 --test rescan behavior of gather merge
 set enable_material = false;
 
@@ -182,12 +168,6 @@ select string4 from tenk1 order by string4 limit 5;
 reset max_parallel_workers;
 reset enable_hashagg;
 
--- check parallelized int8 aggregate (bug #14897)
-explain (costs off)
-select avg(unique1::int8) from tenk1;
-
-select avg(unique1::int8) from tenk1;
-
 -- test the sanity of parallel query after the active role is dropped.
 drop role if exists regress_parallel_worker;
 create role regress_parallel_worker;
@@ -204,36 +184,9 @@ set force_parallel_mode=1;
 explain (costs off)
   select stringu1::int2 from tenk1 where unique1 = 1;
 
--- Window function calculation can't be pushed to workers.
-explain (costs off, verbose)
-  select count(*) from tenk1 a where (unique1, two) in
-    (select unique1, row_number() over() from tenk1 b);
-
-
--- LIMIT/OFFSET within sub-selects can't be pushed to workers.
-explain (costs off)
-  select * from tenk1 a where two in
-    (select two from tenk1 b where stringu1 like '%AAAA' limit 3);
-
 -- to increase the parallel query test coverage
 EXPLAIN (analyze, timing off, summary off, costs off) SELECT * FROM tenk1;
 
--- test passing expanded-value representations to workers
-CREATE FUNCTION make_some_array(int,int) returns int[] as
-$$declare x int[];
-  begin
-    x[1] := $1;
-    x[2] := $2;
-    return x;
-  end$$ language plpgsql parallel safe;
-CREATE TABLE fooarr(f1 text, f2 int[], f3 text);
-INSERT INTO fooarr VALUES('1', ARRAY[1,2], 'one');
-
-PREPARE pstmt(text, int[]) AS SELECT * FROM fooarr WHERE f1 = $1 AND f2 = $2;
-EXPLAIN (COSTS OFF) EXECUTE pstmt('1', make_some_array(1,2));
-EXECUTE pstmt('1', make_some_array(1,2));
-DEALLOCATE pstmt;
-
 -- provoke error in worker
 select stringu1::int2 from tenk1 where unique1 = 1;
 
